--[[
	Copyright (C) Udorn (Blackhand)
	
	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.	
--]]

-- Data found on http://www.wowwiki.com/Disenchanting_tables and adjusted by Wowhead and Echantrix
   
vendor.Disenchant = vendor.Vendor:NewModule("Disenchant")

local L = vendor.Locale.GetInstance()
local log = vendor.Debug:new("Disenchant")

local STRANGE_DUST = 10940
local SOUL_DUST = 11083
local VISION_DUST = 11137
local DREAM_DUST = 11176
local ILLUSION_DUST = 16204
local ARCANE_DUST = 22445
local INFINITE_DUST = 34054
local HYPNOTIC_DUST = 52555

local L_MAGIC_ESSENCE = 10938
local G_MAGIC_ESSENCE = 10939
local L_ASTRAL_ESSENCE = 10998
local G_ASTRAL_ESSENCE = 11082
local L_MYSTIC_ESSENCE = 11134
local G_MYSTIC_ESSENCE = 11135
local L_NETHER_ESSENCE = 11174
local G_NETHER_ESSENCE = 11175
local L_ETERNAL_ESSENCE = 16202
local G_ETERNAL_ESSENCE = 16203
local L_PLANAR_ESSENCE = 22447
local G_PLANAR_ESSENCE = 22446
local L_COSMIC_ESSENCE = 34056
local G_COSMIC_ESSENCE = 34055
local L_CELESTIAL_ESSENCE = 52718
local G_CELESTIAL_ESSENCE = 52719

local S_GLIMMERING_SHARD = 10978
local L_GLIMMERING_SHARD = 11084
local S_GLOWING_SHARD = 11138
local L_GLOWING_SHARD = 11139
local S_RADIANT_SHARD = 11177
local L_RADIANT_SHARD = 11178
local S_BRILLIANT_SHARD = 14343
local L_BRILLIANT_SHARD = 14344
local S_PRISMATIC_SHARD = 22448
local L_PRISMATIC_SHARD = 22449
local S_DREAM_SHARD = 34053
local DREAM_SHARD = 34052
local S_HEAVENLY_SHARD = 52720
local HEAVENLY_SHARD = 52721

local NEXUS_CRYSTAL = 20725
local VOID_CRYSTAL = 22450
local ABYSS_CRYSTAL = 34057
local MAELSTROM_CRYSTAL = 52722

local ITEM_LEVEL = 1
local ENTRIES = 2
local ITEM_ID = 1
local PROXIMITY = 2
local MIN = 3
local MAX = 4
local ESTIMATED = 5

local ENCHANTING_ICON = "Interface\\Icons\\Trade_Engraving"

local EXCEPTIONS = {
	[20408] = true, -- Twilight Cultist Cowl
	[20406] = true, -- Twilight Cultist Mantle
	[20407] = true, -- Twilight Cultist Robe
	[11287] = true, -- Lesser Magic Wand
	[11288] = true, -- Greater Magic Wand
}

-- taken from http://www.wowwiki.com/Disenchanting
local REQUIRED_SKILL = {
	-- uncomon (1)
	{{1,1},{25,21},{50,26},{75,31},{100,36},{125,41},{150,46},{175,51},
	{200,56},{225,61},{275,100},{325,130},{350,154},{425,183},{nil,334}},
	-- rare (2)
	{{1,1,},{25,21},{50,26},{75,31},{100,36},{125,41},{150,46},{175,51},{200,56},
	{225,61},{275,100},{325,130},{450,288},{nil,378}},
	-- epic (3)
	{{225,61},{300,90},{375,200},{475,353},{nil,417}}
}

-- { itemLevel, { {itemId, proximity, min [, max [, estimated] ] }, ...} } 
local DE_INFO_UNCOMMON_ARMOR = {
	{5, {{STRANGE_DUST, 0.8, 1, 2}, {L_MAGIC_ESSENCE, 0.2, 1, 2} } },
	{16, {{STRANGE_DUST, 0.75, 2, 3}, {G_MAGIC_ESSENCE, 0.2, 1, 2}, {S_GLIMMERING_SHARD, 0.05, 1} } },
	{21, {{STRANGE_DUST, 0.75, 4, 6}, {L_ASTRAL_ESSENCE, 0.15, 1, 2}, {S_GLIMMERING_SHARD, 0.1, 1} } },
	{26, {{SOUL_DUST, 0.75, 1, 2}, {G_ASTRAL_ESSENCE, 0.2, 1, 2}, {L_GLIMMERING_SHARD, 0.05, 1} } },
	{31, {{SOUL_DUST, 0.75, 2, 5}, {L_MYSTIC_ESSENCE, 0.2, 1, 2}, {S_GLOWING_SHARD, 0.05, 1} } },
	{36, {{VISION_DUST, 0.75, 1, 2}, {G_MYSTIC_ESSENCE, 0.2, 1, 2}, {L_GLOWING_SHARD, 0.05, 1} } },
	{41, {{VISION_DUST, 0.75, 2, 5}, {L_NETHER_ESSENCE, 0.2, 1, 2}, {S_RADIANT_SHARD, 0.05, 1} } },
	{46, {{DREAM_DUST, 0.75, 1, 2}, {G_NETHER_ESSENCE, 0.2, 1, 2}, {L_RADIANT_SHARD, 0.05, 1} } },
	{51, {{DREAM_DUST, 0.75, 2, 5}, {L_ETERNAL_ESSENCE, 0.2, 1, 2}, {S_BRILLIANT_SHARD, 0.05, 1} } },
	{56, {{ILLUSION_DUST, 0.75, 1, 2}, {G_ETERNAL_ESSENCE, 0.2, 1, 2}, {L_BRILLIANT_SHARD, 0.05, 1} } },
	{61, {{ILLUSION_DUST, 0.75, 2, 5}, {G_ETERNAL_ESSENCE, 0.2, 2, 3}, {L_BRILLIANT_SHARD, 0.05, 1} } },
	{79, {{ARCANE_DUST, 0.75, 1, 3}, {L_PLANAR_ESSENCE, 0.22, 1, 3}, {S_PRISMATIC_SHARD, 0.03, 1} } },
	{80, {{ARCANE_DUST, 0.75, 2, 3}, {L_PLANAR_ESSENCE, 0.22, 2, 3}, {S_PRISMATIC_SHARD, 0.03, 1} } },
	{100, {{ARCANE_DUST, 0.75, 2, 5}, {G_PLANAR_ESSENCE, 0.22, 1, 2}, {L_PRISMATIC_SHARD, 0.03, 1} } },
	{130, {{INFINITE_DUST, 0.75, 1, 3}, {L_COSMIC_ESSENCE, 0.22, 1, 2}, {S_DREAM_SHARD, 0.03, 1}}},
	{152, {{INFINITE_DUST, 0.75, 4, 7}, {G_COSMIC_ESSENCE, 0.22, 1, 2}, {DREAM_SHARD, 0.03, 1}}},
	{278, {{HYPNOTIC_DUST, 0.75, 1, 3}, {L_CELESTIAL_ESSENCE, 0.25, 1, 3}}},
	{306, {{HYPNOTIC_DUST, 0.735, 1, 5}, {G_CELESTIAL_ESSENCE, 0.25, 1, 2}, {L_CELESTIAL_ESSENCE, 0.015, 1, 3}}},
	{316, {{HYPNOTIC_DUST, 0.75, 1, 6}, {G_CELESTIAL_ESSENCE, 0.25, 1, 4}}}
}

local DE_INFO_UNCOMMON_WEAPONS = {
	{6, {{L_MAGIC_ESSENCE, 0.8, 1, 2}, {STRANGE_DUST, 0.2, 1, 2}}},
	{16, {{G_MAGIC_ESSENCE, 0.75, 1, 2}, {STRANGE_DUST, 0.2, 2, 3}, {S_GLIMMERING_SHARD, 0.05, 1}}},
	{21, {{L_ASTRAL_ESSENCE, 0.75, 1, 2}, {STRANGE_DUST, 0.15, 4, 6}, {S_GLIMMERING_SHARD, 0.1, 1}}},
	{26, {{G_ASTRAL_ESSENCE, 0.75, 1, 2}, {SOUL_DUST, 0.2, 1, 2}, {L_GLIMMERING_SHARD, 0.05, 1}}},
	{31, {{L_MYSTIC_ESSENCE, 0.75, 1, 2}, {SOUL_DUST, 0.2, 2, 5}, {S_GLOWING_SHARD, 0.05, 1}}},
	{36, {{G_MYSTIC_ESSENCE, 0.75, 1, 2}, {VISION_DUST, 0.2, 1, 2}, {L_GLOWING_SHARD, 0.05, 1}}},
	{41, {{L_NETHER_ESSENCE, 0.75, 1, 2}, {VISION_DUST, 0.2, 2, 5}, {S_RADIANT_SHARD, 0.05, 1}}},
	{46, {{G_NETHER_ESSENCE, 0.75, 1, 2}, {DREAM_DUST, 0.2, 1, 2}, {L_RADIANT_SHARD, 0.05, 1}}},
	{51, {{L_ETERNAL_ESSENCE, 0.75, 1, 2}, {DREAM_DUST, 0.22, 2, 5}, {S_BRILLIANT_SHARD, 0.03, 1}}},
	{56, {{G_ETERNAL_ESSENCE, 0.75, 1, 2}, {ILLUSION_DUST, 0.22, 1, 2}, {L_BRILLIANT_SHARD, 0.03, 1}}},
	{61, {{G_ETERNAL_ESSENCE, 0.75, 2, 3}, {ILLUSION_DUST, 0.22, 2, 5}, {L_BRILLIANT_SHARD, 0.03, 1}}},
	{80, {{L_PLANAR_ESSENCE, 0.75, 2, 3}, {ARCANE_DUST, 0.22, 2, 3}, {S_PRISMATIC_SHARD, 0.03, 1}}},
	{100, {{G_PLANAR_ESSENCE, 0.75, 1, 2}, {ARCANE_DUST, 0.22, 2, 5}, {L_PRISMATIC_SHARD, 0.03, 1}}},
	{130, {{L_COSMIC_ESSENCE, 0.75, 1, 2}, {INFINITE_DUST, 0.22, 1, 4}, {S_DREAM_SHARD, 0.03, 1}}},
	{152, {{G_COSMIC_ESSENCE, 0.75, 1, 2}, {INFINITE_DUST, 0.22, 4, 7}, {DREAM_SHARD, 0.03, 1}}},
	{272, {{L_CELESTIAL_ESSENCE, 0.75, 1, 3}, {HYPNOTIC_DUST, 0.25, 1, 3}}},
	{289, {{L_CELESTIAL_ESSENCE, 0.75, 1, 4}, {HYPNOTIC_DUST, 0.25, 1, 4}}},
	{306, {{G_CELESTIAL_ESSENCE, 0.75, 1, 2}, {HYPNOTIC_DUST, 0.25, 1, 5}}},
	{318, {{G_CELESTIAL_ESSENCE, 0.75, 2, 4}, {HYPNOTIC_DUST, 0.25, 1, 7}}}
}

local DE_INFO_RARE = {
	{1, {{S_GLIMMERING_SHARD, 1, 1}}},
	{26, {{L_GLIMMERING_SHARD, 1, 1}}},
	{31, {{S_GLOWING_SHARD, 1, 1}}},
	{36, {{L_GLOWING_SHARD, 1, 1}}},
	{41, {{S_RADIANT_SHARD, 1, 1}}},
	{46, {{L_RADIANT_SHARD, 1, 1}}},
	{51, {{S_BRILLIANT_SHARD, 1, 1}}},
	{56, {{L_BRILLIANT_SHARD, 1, 1}}},
	{66, {{S_PRISMATIC_SHARD, 1, 1}}},
	{100, {{L_PRISMATIC_SHARD, 1, 1}}},
	{130, {{S_DREAM_SHARD, 1, 1}}},
	{165, {{DREAM_SHARD, 1, 1}}},
	{292, {{S_HEAVENLY_SHARD, 1, 1}}},
	{325, {{HEAVENLY_SHARD, 1, 1}}}
}

local DE_INFO_EPIC = {
	{40, {{S_RADIANT_SHARD, 1, 2, 4}}},
	{46, {{L_RADIANT_SHARD, 1, 2, 4}}},
	{51, {{S_BRILLIANT_SHARD, 1, 2, 4}}},
	{56, {{NEXUS_CRYSTAL, 1, 1}}},
	{61, {{NEXUS_CRYSTAL, 1, 1}}},
	{95, {{VOID_CRYSTAL, 1, 1}}},
	{105, {{VOID_CRYSTAL, 1, 1}}},
	{165, {{ABYSS_CRYSTAL, 1, 1}}},
	{201, {{ABYSS_CRYSTAL, 1, 1}}},
	{300, {{MAELSTROM_CRYSTAL, 1, 1}}},
}

local function _GetPlayerSkill(self)
	local rtn
	local name
	local prof1, prof2 = GetProfessions()
	local icon, skillLevel
	if (prof1) then
		name, icon, skillLevel = GetProfessionInfo(prof1)
		if (ENCHANTING_ICON == icon) then
			rtn = skillLevel
		end
	end
	if (not rtn and prof2) then
		name, icon, skillLevel = GetProfessionInfo(prof2)
		if (ENCHANTING_ICON == icon) then
			rtn = skillLevel
		end
	end
	return rtn, name
end

local function _GetRequiredSkill(itemLevel, itemRarity)
	local table
	if (itemRarity >= 2 and itemRarity <= 4) then
		table = REQUIRED_SKILL[itemRarity - 1]
	end
	if (table) then
		local n = #table
		local minLevel, maxLevel
		local left, right, idx = 1, n, 0
	  	while (left <= right) do
	  		idx = math.floor((left + right) / 2)
	  		minLevel = table[idx][2]
	  		if (idx < n) then
	  			maxLevel = table[idx + 1][2] - 1
	  		else
	  			maxLevel = 999999
	  		end
	  		if (itemLevel >= minLevel and itemLevel <= maxLevel) then
	  			return table[idx][1]
	  		elseif (itemLevel < minLevel) then
	     		-- search left
	 			right = idx - 1
	  		else
	 			-- search right
	     		left = idx + 1
	  		end
		end
		return table[left][1]
	end
	return nil
end

local function _ValidateTable(table)
	local n = #table
	for i = 1,n do
		local info = table[i]
		if (not info[ITEM_LEVEL] or type(info[ITEM_LEVEL]) ~= "number") then
			log:Debug("Illegal level for entry %s", i)
		end
		if (not info[ENTRIES] or type(info[ENTRIES]) ~= "table") then
			log:Debug("Illegal table for entry %s", i)
		end
		local proximity = 0
		for k,v in pairs(info[ENTRIES]) do
			if (not v[ITEM_ID] or type(v[ITEM_ID]) ~= "number") then
				log:Debug("Illegal itemId for entry %s", i)
			end
			if (not v[MIN] or type(v[MIN]) ~= "number") then
				log:Debug("Illegal min for entry %s", i)
			end
			if (not v[PROXIMITY] or type(v[PROXIMITY]) ~= "number") then
				log:Debug("Illegal proximity for entry %s", i)
			end
			proximity = proximity + v[PROXIMITY]
		end
		if (proximity < 0.9999999 or proximity > 1.000001) then
			log:Debug("Illegal proximity %s for entry %s", proximity, i)
		end
	end
end

local function _FindDisenchantEntry(table, itemLevel)
	local n = #table
	local left, right, idx = 1, n, 0
  	local cmp
  	while (left <= right) do
  		idx = math.floor((left + right) / 2)
   		cmp = table[idx][ITEM_LEVEL]
     	if (itemLevel == cmp) then
     		return idx
     	elseif (itemLevel < cmp) then
     		-- search left
 			right = idx - 1
     	else
 			-- search right
     		left = idx + 1
     	end
	end
	if (idx >= 1 and idx <= n) then
		if (itemLevel < table[idx][ITEM_LEVEL]) then
			if (idx > 1) then
				idx = idx - 1
			else
				idx = nil
			end
		end
	else
		idx = nil
	end
	return idx
end

local function _GetDisenchantMessage(entry)
	local texture = "|T"..GetItemIcon(entry[ITEM_ID])..":14|t"
	local name = GetItemInfo(entry[ITEM_ID]) or ""
	if (entry[MAX]) then
		return ((entry[MIN] + entry[MAX]) / 2.0).." "..texture.." "..name.." "..(entry[PROXIMITY] * 100).."%"
	end
	return entry[MIN].." "..texture.." "..name.." "..(entry[PROXIMITY] * 100).."%"
end

local function _GetDisenchantInfo(self, table, itemLevel)
	local idx = _FindDisenchantEntry(table, itemLevel)
	if (idx) then
		local entries = table[idx][ENTRIES]
		for k,v in pairs(entries) do
			log:Debug(_GetDisenchantMessage(v))
		end
	end
end

local function _GetDisenchantTable(self, itemLink)
	local table = nil
	local itemName, _, itemRarity, itemLevel, _, itemType = GetItemInfo(itemLink)
	if (itemName and (itemType == self.armor or itemType == self.weapons)) then
		if (itemRarity == 2) then
			if (itemType == self.armor) then
				table = DE_INFO_UNCOMMON_ARMOR
			elseif (itemType == self.weapons) then
				table = DE_INFO_UNCOMMON_WEAPONS
			end
		elseif (itemRarity == 3) then
			table = DE_INFO_RARE
		elseif (itemRarity == 4) then
			table = DE_INFO_EPIC
		end
	end 
	return table, itemLevel, itemRarity
end

function vendor.Disenchant:OnEnable()
	-- get localized versions of Armor and Weapon
	local classes = {GetAuctionItemClasses()}
	self.weapons = classes[1]
	self.armor = classes[2]
	
	_ValidateTable(DE_INFO_UNCOMMON_ARMOR)
	_ValidateTable(DE_INFO_UNCOMMON_WEAPONS)
	_ValidateTable(DE_INFO_RARE)
	_ValidateTable(DE_INFO_EPIC)
	
	vendor.TooltipHook:Enable()
	vendor.TooltipHook:AddAppender(self, 100)
end

function vendor.Disenchant:AppendToGameTooltip(tooltip, itemLink, itemName, count)

	if (not vendor.TooltipHook.db.profile.showDisenchant) then
		return
	end
	
	local id = vendor.Items:GetItemLinkId(itemLink)
--	log:Debug("id %s", id)
	if (id and EXCEPTIONS[tonumber(id)]) then
		return
	end
	
	local table, itemLevel, itemRarity = _GetDisenchantTable(self, itemLink)
	if (table and itemLevel) then
		local idx = _FindDisenchantEntry(table, itemLevel)
		if (idx) then
			local expectation = 0
			local lowerExpectation = 0
			local entries = table[idx][ENTRIES]
			for k,v in pairs(entries) do
				local msg = _GetDisenchantMessage(v)
				tooltip:AddLine(msg);
				
				-- TODO switch to itemId in backend
				local ilink = vendor.Items:GetItemLink(v[ITEM_ID]) 				
				local _, avgBuyout, _, lowerBuyout = vendor.Gatherer:GetCurrentAuctionInfo(ilink)
				if (not avgBuyout or avgBuyout == 0) then
					avgBuyout = select(2, vendor.Gatherer:GetAuctionInfo(ilink))
				end
				if (avgBuyout and avgBuyout > 0) then
					local q = v[MIN]
					if (v[ESTIMATED]) then
						q = v[ESTIMATED]
					elseif (v[MAX]) then
						q = (v[MIN] + v[MAX]) / 2
					end
					expectation = expectation + (avgBuyout * q * v[PROXIMITY])
					if (lowerBuyout and lowerBuyout > 0 and lowerExpectation) then
						lowerExpectation = lowerExpectation + (lowerBuyout * q * v[PROXIMITY])
					else
						lowerExpectation = nil
					end
				end
			end
			if (lowerExpectation and lowerExpectation > 0) then
				local msg1 = L["Disenchant value"]
				local msg2 = vendor.Format.FormatMoney(lowerExpectation, true)
				tooltip:AddDoubleLine(msg1, msg2)
			elseif (expectation > 0) then
				local msg1 = L["Disenchant value"]
				local msg2 = vendor.Format.FormatMoney(expectation, true)
				tooltip:AddDoubleLine(msg1, msg2)
			end
			
			local requiredSkill = _GetRequiredSkill(itemLevel, itemRarity)
			if (requiredSkill) then
				local playerSkill, profName = _GetPlayerSkill(self)
				if (not playerSkill or playerSkill < requiredSkill) then
					tooltip:AddLine("|cffff0000"..ITEM_DISENCHANT_MIN_SKILL:format(profName, requiredSkill).."|r")
				end
			end
		end
	end
end

function vendor.Disenchant:GetDisenchantValue(itemLink)
	local id = vendor.Items:GetItemLinkId(itemLink)
	if (id and EXCEPTIONS[tonumber(id)]) then
		return 0
	end
	
	local rtn = 0
	local table, itemLevel, itemRarity = _GetDisenchantTable(self, itemLink)
	if (table and itemLevel) then
		local idx = _FindDisenchantEntry(table, itemLevel)
		if (idx) then
			local expectation = 0
			local lowerExpectation = 0
			local entries = table[idx][ENTRIES]
			for k,v in pairs(entries) do
				-- TODO switch to itemId in backend
				local ilink = vendor.Items:GetItemLink(v[ITEM_ID]) 				
				local _, avgBuyout, _, lowerBuyout = vendor.Gatherer:GetCurrentAuctionInfo(ilink)
				if (not avgBuyout or avgBuyout == 0) then
					avgBuyout = select(2, vendor.Gatherer:GetAuctionInfo(ilink))
				end
				if (avgBuyout and avgBuyout > 0) then
					local q = v[MIN]
					if (v[ESTIMATED]) then
						q = v[ESTIMATED]
					elseif (v[MAX]) then
						q = (v[MIN] + v[MAX]) / 2
					end
					expectation = expectation + (avgBuyout * q * v[PROXIMITY])
					if (lowerBuyout and lowerBuyout > 0 and lowerExpectation) then
						lowerExpectation = lowerExpectation + (lowerBuyout * q * v[PROXIMITY])
					else
						lowerExpectation = nil
					end
				end
			end
			if (lowerExpectation and lowerExpectation > 0) then
				rtn = lowerExpectation
			else
				rtn = expectation
			end
		end
	end
	return rtn
end